@@ -140,7 +140,7 @@ static const char* app_loader_template =
 ;
 
 static const char *output_file_name;
-static const char *output_debug_file;
+static struct strarray output_debug_files;
 static const char *output_implib;
 static const char *output;
 static int keep_generated = 0;
@@ -170,7 +170,6 @@ enum file_type { file_na, file_other, file_obj, file_res, file_rc, file_arh, fil
 static bool is_pe;
 static bool is_static;
 static bool is_shared;
-static bool is_gui_app;
 static bool is_unicode_app;
 static bool is_win16_app;
 static bool is_arm64x;
@@ -208,8 +207,8 @@ static struct strarray delayimports;
 static void cleanup_output_files(void)
 {
     if (output_file_name) unlink( output_file_name );
-    if (output_debug_file) unlink( output_debug_file );
     if (output_implib) unlink( output_implib );
+    STRARRAY_FOR_EACH( file, &output_debug_files) unlink( file );
 }
 
 static void clean_temp_files(void)
@@ -341,11 +340,9 @@ static enum file_type guess_lib_type(const char* dir, const char* library, char*
 
 static enum file_type get_lib_type(struct strarray path, const char *library, char** file)
 {
-    unsigned int i;
-
-    for (i = 0; i < path.count; i++)
+    STRARRAY_FOR_EACH( str, &path )
     {
-        enum file_type type = guess_lib_type(path.str[i], library, file);
+        enum file_type type = guess_lib_type(str, library, file);
 	if (type != file_na) return type;
     }
     return file_na;
@@ -355,7 +352,6 @@ static const char *find_binary( const char *name )
 {
     char *file_name, *args;
     struct strarray dirs = empty_strarray;
-    unsigned int i;
 
     if (strchr( name, '/' )) return name;
 
@@ -364,10 +360,10 @@ static const char *find_binary( const char *name )
 
     strarray_addall( &dirs, prefix_dirs );
     strarray_addall( &dirs, path_dirs );
-    for (i = 0; i < dirs.count; i++)
+    STRARRAY_FOR_EACH( dir, &dirs )
     {
         struct stat st;
-        char *prog = strmake( "%s/%s%s", dirs.str[i], file_name, EXEEXT );
+        char *prog = strmake( "%s/%s%s", dir, file_name, EXEEXT );
         if (stat( prog, &st ) == 0 && S_ISREG( st.st_mode ) && (st.st_mode & 0111))
             return args ? strmake( "%s %s", prog, args ) : prog;
         free( prog );
@@ -499,6 +495,7 @@ static struct strarray get_link_args( const char *output_name )
 {
     struct strarray link_args = get_translator();
     struct strarray flags = empty_strarray;
+    char *version;
 
     strarray_addall( &link_args, linker_args );
 
@@ -544,10 +541,9 @@ static struct strarray get_link_args( const char *output_name )
             strarray_add( &flags, "-shared" );
             strarray_add( &flags, "-Wl,--kill-at" );
         }
-        else strarray_add( &flags, is_gui_app ? "-mwindows" : "-mconsole" );
 
         if (is_unicode_app) strarray_add( &flags, "-municode" );
-        if (subsystem) strarray_add( &flags, strmake("-Wl,--subsystem,%s", subsystem ));
+        strarray_add( &flags, strmake( "-Wl,--subsystem,%s", subsystem ) );
 
         strarray_add( &flags, "-Wl,--exclude-all-symbols" );
         strarray_add( &flags, "-Wl,--nxcompat" );
@@ -562,8 +558,8 @@ static struct strarray get_link_args( const char *output_name )
         if (entry_point)
             strarray_add( &flags, strmake( "-Wl,--entry,%s%s", target.cpu == CPU_i386 ? "_" : "", entry_point ));
 
-        if (output_debug_file && strendswith(output_debug_file, ".pdb"))
-            strarray_add(&link_args, strmake("-Wl,--pdb=%s", output_debug_file));
+        STRARRAY_FOR_EACH( file, &output_debug_files )
+            if (strendswith(file, ".pdb")) strarray_add(&link_args, strmake("-Wl,--pdb=%s", file));
 
         if (use_build_id)
             strarray_add( &link_args, "-Wl,--build-id");
@@ -597,22 +593,30 @@ static struct strarray get_link_args( const char *output_name )
         if (is_unicode_app) strarray_add( &flags, "-municode" );
         if (nostartfiles) strarray_add( &flags, "-nostartfiles" );
         if (image_base) strarray_add( &flags, strmake("-Wl,-base:%s", image_base ));
+
+        if (large_address_aware && target.cpu == CPU_i386)
+            strarray_add( &flags, "-Wl,-largeaddressaware" );
+
         if (entry_point) strarray_add( &flags, strmake( "-Wl,-entry:%s", entry_point ));
 
-        if (subsystem)
-            strarray_add( &flags, strmake("-Wl,-subsystem:%s", subsystem ));
-        else
-            strarray_add( &flags, strmake("-Wl,-subsystem:%s", is_gui_app ? "windows" : "console" ));
+        if ((version = strchr( subsystem, ':' ))) subsystem = strmake( "%.*s,%s", (int)(version - subsystem), subsystem, version + 1 );
+        strarray_add( &flags, "-Xlinker" );
+        strarray_add( &flags, strmake( "-subsystem:%s", subsystem ) );
 
-        if (output_debug_file && strendswith(output_debug_file, ".pdb"))
+        STRARRAY_FOR_EACH( file, &output_debug_files )
         {
-            strarray_add(&link_args, "-Wl,-debug");
-            strarray_add(&link_args, strmake("-Wl,-pdb:%s", output_debug_file));
+            if (strendswith(file, ".pdb"))
+            {
+                strarray_add(&link_args, "-Wl,-debug");
+                strarray_add(&link_args, strmake("-Wl,-pdb:%s", file));
+            }
+            else strarray_add( &link_args, "-Wl,-debug:dwarf" );
         }
-        else if (strip)
+
+        if (strip)
             strarray_add( &link_args, "-s" );
-        else
-            strarray_add(&link_args, "-Wl,-debug:dwarf");
+        else if (!output_debug_files.count)
+            strarray_add( &link_args, "-Wl,-debug:dwarf" );
 
         if (use_build_id)
             strarray_add( &link_args, "-Wl,-build-id");
@@ -855,7 +859,7 @@ static struct strarray get_compat_defines( int gcc_defs )
 static void compile( struct strarray files, const char *output_name, int compile_only )
 {
     struct strarray comp_args = get_translator();
-    unsigned int i, j;
+    unsigned int j;
     int gcc_defs = 0;
     struct strarray gcc;
     struct strarray gpp;
@@ -871,14 +875,14 @@ static void compile( struct strarray files, const char *output_name, int compile
 	case proc_cxx:
             gcc = build_tool_name( target_alias, tool_cc );
             gpp = build_tool_name( target_alias, tool_cxx );
-            for ( j = 0; !gcc_defs && j < comp_args.count; j++ )
+            STRARRAY_FOR_EACH( cc, &comp_args )
             {
-                const char *cc = comp_args.str[j];
-
-                for (i = 0; !gcc_defs && i < gcc.count; i++)
-                    gcc_defs = gcc.str[i][0] != '-' && strendswith(cc, gcc.str[i]);
-                for (i = 0; !gcc_defs && i < gpp.count; i++)
-                    gcc_defs = gpp.str[i][0] != '-' && strendswith(cc, gpp.str[i]);
+                STRARRAY_FOR_EACH( str, &gcc )
+                    if ((gcc_defs = str[0] != '-' && strendswith(cc, str))) break;
+                if (gcc_defs) break;
+                STRARRAY_FOR_EACH( str, &gpp )
+                    if ((gcc_defs = str[0] != '-' && strendswith(cc, str))) break;
+                if (gcc_defs) break;
             }
             break;
     }
@@ -902,14 +906,14 @@ static void compile( struct strarray files, const char *output_name, int compile
     strarray_addall(&comp_args, compiler_args);
 
     /* last, but not least, the files */
-    for ( j = 0; j < files.count; j++ )
+    STRARRAY_FOR_EACH( file, &files )
     {
-        if (files.str[j][0] == '-')
+        if (file[0] == '-')
         {
             /* keep -x and bare '-' (i.e. stdin) options */
-            if (files.str[j][1] && files.str[j][1] != 'x') continue;
+            if (file[1] && file[1] != 'x') continue;
         }
-	strarray_add(&comp_args, files.str[j]);
+	strarray_add(&comp_args, file);
     }
 
     /* standard includes come last in the include search path */
@@ -964,7 +968,6 @@ static struct strarray get_winebuild_args( const char *target )
 {
     const char *binary;
     struct strarray spec_args = empty_strarray;
-    unsigned int i;
 
     if (!(binary = find_binary( winebuild ))) error( "Could not find winebuild\n" );
     strarray_add( &spec_args, binary );
@@ -977,8 +980,8 @@ static struct strarray get_winebuild_args( const char *target )
     }
     if (force_pointer_size)
         strarray_add(&spec_args, strmake("-m%u", 8 * force_pointer_size ));
-    for (i = 0; i < prefix_dirs.count; i++)
-        strarray_add( &spec_args, strmake( "-B%s", prefix_dirs.str[i] ));
+    STRARRAY_FOR_EACH( dir, &prefix_dirs )
+        strarray_add( &spec_args, strmake( "-B%s", dir ));
     strarray_addall( &spec_args, winebuild_args );
     return spec_args;
 }
@@ -1015,15 +1018,14 @@ static const char *find_libgcc(void)
     const char *out = make_temp_file( "find_libgcc", ".out" );
     const char *err = make_temp_file( "find_libgcc", ".err" );
     struct strarray link = get_translator();
-    int sout = -1, serr = -1, i;
+    int sout = -1, serr = -1;
     char *libgcc, *p;
     struct stat st;
     size_t cnt;
     int ret;
 
-    for (i = 0; i < linker_args.count; i++)
-	if (strcmp(linker_args.str[i], "--no-default-config" ))
-            strarray_add( &link, linker_args.str[i] );
+    STRARRAY_FOR_EACH( arg, &linker_args )
+	if (strcmp(arg, "--no-default-config" )) strarray_add( &link, arg );
 
     strarray_add( &link, "-print-libgcc-file-name" );
 
@@ -1088,7 +1090,6 @@ static void build_spec_obj( const char *spec_file, const char *output_file,
                             const char *target_name, struct strarray files,
                             struct strarray resources, struct strarray *spec_objs )
 {
-    unsigned int i;
     struct strarray spec_args = get_winebuild_args( target_name );
     struct strarray tool;
     const char *spec_o_name, *output_name;
@@ -1134,12 +1135,10 @@ static void build_spec_obj( const char *spec_file, const char *output_file,
         strarray_add(&spec_args, output_name);
     }
 
-    if (!is_shared)
-    {
-        strarray_add(&spec_args, "--subsystem");
-        strarray_add(&spec_args, is_gui_app ? "windows" : "console");
-        if (large_address_aware) strarray_add( &spec_args, "--large-address-aware" );
-    }
+    strarray_add( &spec_args, "--subsystem" );
+    strarray_add( &spec_args, subsystem );
+
+    if (!is_shared && large_address_aware) strarray_add( &spec_args, "--large-address-aware" );
 
     if (target.platform == PLATFORM_WINDOWS && target.cpu == CPU_i386)
         strarray_add(&spec_args, "--safeseh");
@@ -1150,30 +1149,20 @@ static void build_spec_obj( const char *spec_file, const char *output_file,
         strarray_add(&spec_args, entry_point);
     }
 
-    if (subsystem)
-    {
-        strarray_add(&spec_args, "--subsystem");
-        strarray_add(&spec_args, subsystem);
-    }
-
-    if (!is_pe)
-    {
-        for (i = 0; i < delayimports.count; i++)
-            strarray_add(&spec_args, strmake("-d%s", delayimports.str[i]));
-    }
+    if (!is_pe) STRARRAY_FOR_EACH( imp, &delayimports ) strarray_add(&spec_args, strmake("-d%s", imp));
 
     strarray_addall( &spec_args, resources );
 
     /* add other files */
     strarray_add(&spec_args, "--");
-    for (i = 0; i < files.count; i++)
+    STRARRAY_FOR_EACH( file, &files )
     {
-	switch(files.str[i][1])
+	switch(file[1])
 	{
 	    case 'd':
 	    case 'a':
 	    case 'o':
-		strarray_add(&spec_args, files.str[i] + 2);
+		strarray_add(&spec_args, file + 2);
 		break;
 	}
     }
@@ -1185,7 +1174,6 @@ static void build_spec_obj( const char *spec_file, const char *output_file,
 /* run winebuild to generate a data-only library */
 static void build_data_lib( const char *spec_file, const char *output_file, struct strarray files )
 {
-    unsigned int i;
     struct strarray spec_args = get_winebuild_args( target_alias );
 
     strarray_add(&spec_args, is_shared ? "--dll" : "--exe");
@@ -1198,8 +1186,7 @@ static void build_data_lib( const char *spec_file, const char *output_file, stru
     }
 
     /* add resource files */
-    for (i = 0; i < files.count; i++)
-	if (files.str[i][1] == 'r') strarray_add(&spec_args, files.str[i]);
+    STRARRAY_FOR_EACH( file, &files ) if (file[1] == 'r') strarray_add(&spec_args, file);
 
     spawn(spec_args, 0);
 }
@@ -1216,7 +1203,6 @@ static void build(struct strarray input_files, const char *output)
     const char *libgcc = NULL;
     int generate_app_loader = 1;
     const char *crt_lib = NULL;
-    unsigned int i, j;
 
     /* NOTE: for the files array we'll use the following convention:
      *    -axxx:  xxx is an archive (.a)
@@ -1269,9 +1255,8 @@ static void build(struct strarray input_files, const char *output)
 
     /* mark the files with their appropriate type */
     spec_file = lang = 0;
-    for ( j = 0; j < input_files.count; j++ )
+    STRARRAY_FOR_EACH( file, &input_files )
     {
-	const char* file = input_files.str[j];
 	if (file[0] != '-')
 	{
 	    switch(get_file_type(file))
@@ -1323,7 +1308,7 @@ static void build(struct strarray input_files, const char *output)
 
     if (!wine_objdir && !nodefaultlibs)
     {
-        if (is_gui_app)
+        if (!strncmp( subsystem, "windows", 7 ))
 	{
 	    add_library(lib_dirs, &files, "shell32");
 	    add_library(lib_dirs, &files, "comdlg32");
@@ -1353,7 +1338,7 @@ static void build(struct strarray input_files, const char *output)
     /* set default entry point, if needed */
     if (!entry_point)
     {
-        if (subsystem && !strcmp( subsystem, "native" ))
+        if (!strcmp( subsystem, "native" ))
             entry_point = (is_pe && target.cpu == CPU_i386) ? "DriverEntry@8" : "DriverEntry";
         else if (use_msvcrt && !is_shared && !is_win16_app)
             entry_point = is_unicode_app ? "wmainCRTStartup" : "mainCRTStartup";
@@ -1366,8 +1351,7 @@ static void build(struct strarray input_files, const char *output)
         return;
     }
 
-    for (i = 0; i < files.count; i++)
-	if (files.str[i][1] == 'r') strarray_add( &resources, files.str[i] );
+    STRARRAY_FOR_EACH( file, &files ) if (file[1] == 'r') strarray_add( &resources, file );
 
     build_spec_obj( spec_file, output_file, target_alias, files, resources, &spec_objs );
     if (is_arm64x)
@@ -1400,26 +1384,25 @@ static void build(struct strarray input_files, const char *output)
     strarray_add(&link_args, "-o");
     strarray_add(&link_args, output_file_name);
 
-    for ( j = 0; j < lib_dirs.count; j++ )
-	strarray_add(&link_args, strmake("-L%s", lib_dirs.str[j]));
+    STRARRAY_FOR_EACH( dir, &lib_dirs ) strarray_add(&link_args, strmake("-L%s", dir));
 
     strarray_addall( &link_args, spec_objs );
 
     if (is_pe)
     {
-        for (j = 0; j < delayimports.count; j++)
+        STRARRAY_FOR_EACH( imp, &delayimports )
         {
             if (target.platform == PLATFORM_WINDOWS)
-                strarray_add(&link_args, strmake("-Wl,-delayload:%s", delayimports.str[j]));
+                strarray_add(&link_args, strmake("-Wl,-delayload:%s", imp));
             else
-                strarray_add(&link_args, strmake("-Wl,-delayload,%s",delayimports.str[j]));
+                strarray_add(&link_args, strmake("-Wl,-delayload,%s",imp));
         }
     }
 
-    for ( j = 0; j < files.count; j++ )
+    STRARRAY_FOR_EACH( file, &files )
     {
-	const char* name = files.str[j] + 2;
-	switch(files.str[j][1])
+	const char* name = file + 2;
+	switch(file[1])
 	{
 	    case 'l':
 		strarray_add(&link_args, strmake("-l%s", name));
@@ -1441,7 +1424,7 @@ static void build(struct strarray input_files, const char *output)
 		strarray_add(&link_args, name);
 		break;
 	    case 'W':
-		strarray_add(&link_args, files.str[j]);
+		strarray_add(&link_args, file);
 		break;
 	}
     }
@@ -1459,15 +1442,18 @@ static void build(struct strarray input_files, const char *output)
 
     spawn(link_args, 0);
 
-    if (output_debug_file && !strendswith(output_debug_file, ".pdb"))
+    STRARRAY_FOR_EACH( file, &output_debug_files )
     {
-        struct strarray tool, objcopy = build_tool_name(target_alias, tool_objcopy);
+        struct strarray tool, objcopy;
+
+        if (strendswith(file, ".pdb")) continue;
+        objcopy = build_tool_name(target_alias, tool_objcopy);
 
         tool = empty_strarray;
         strarray_addall( &tool, objcopy );
         strarray_add(&tool, "--only-keep-debug");
         strarray_add(&tool, output_file_name);
-        strarray_add(&tool, output_debug_file);
+        strarray_add(&tool, file);
         spawn(tool, 1);
 
         tool = empty_strarray;
@@ -1479,7 +1465,7 @@ static void build(struct strarray input_files, const char *output)
         tool = empty_strarray;
         strarray_addall( &tool, objcopy );
         strarray_add(&tool, "--add-gnu-debuglink");
-        strarray_add(&tool, output_debug_file);
+        strarray_add(&tool, file);
         strarray_add(&tool, output_file_name);
         spawn(tool, 0);
     }
@@ -1626,6 +1612,7 @@ int main(int argc, char **argv)
     int i, c, next_is_arg = 0;
     int raw_compiler_arg, raw_linker_arg, raw_winebuild_arg;
     struct strarray args = empty_strarray;
+    bool is_gui_app = false;
     const char* option_arg;
     char* str;
 
@@ -1930,7 +1917,7 @@ int main(int argc, char **argv)
                             }
                             if (!strcmp(Wl.str[j], "--debug-file") && j < Wl.count - 1)
                             {
-                                output_debug_file = xstrdup( Wl.str[++j] );
+                                strarray_add( &output_debug_files, xstrdup( Wl.str[++j] ));
                                 continue;
                             }
                             if (!strcmp(Wl.str[j], "--whole-archive") ||
@@ -1958,13 +1945,12 @@ int main(int argc, char **argv)
                     }
 		    else if (strncmp("-Wb,", args.str[i], 4) == 0)
 		    {
-                        unsigned int j;
                         struct strarray Wb = strarray_fromstring(args.str[i] + 4, ",");
-                        for (j = 0; j < Wb.count; j++)
+                        STRARRAY_FOR_EACH( arg, &Wb )
                         {
-                            if (!strcmp(Wb.str[j], "--data-only")) data_only = true;
-                            if (!strcmp(Wb.str[j], "--fake-module")) fake_module = true;
-                            else strarray_add( &winebuild_args, Wb.str[j] );
+                            if (!strcmp(arg, "--data-only")) data_only = true;
+                            if (!strcmp(arg, "--fake-module")) fake_module = true;
+                            else strarray_add( &winebuild_args, arg );
                         }
                         raw_compiler_arg = raw_linker_arg = 0;
 		    }
@@ -2054,6 +2040,14 @@ int main(int argc, char **argv)
 
     if (!file_align) file_align = section_align;
 
+    if (!is_pe && target.cpu != CPU_i386 && target.cpu != CPU_x86_64)
+        error( "Non-PE builds are not supported on this platform. You need to use something like '--target=%s-windows'.\n",
+               target.cpu == CPU_ARM ? "arm" : "aarch64" );
+
+    if (!subsystem) subsystem = is_gui_app ? "windows" : "console";
+    if (!strcmp( subsystem, "windows" ) || !strcmp( subsystem, "console" ))
+        subsystem = strmake( "%s:6.0", subsystem );
+
     if (!winebuild)
     {
         if (wine_objdir) winebuild = strmake( "%s/tools/winebuild/winebuild%s", wine_objdir, EXEEXT );
@@ -2068,7 +2062,7 @@ int main(int argc, char **argv)
     else compile(file_args, output, compile_only);
 
     output_file_name = NULL;
-    output_debug_file = NULL;
+    output_debug_files.count = 0;
     output_implib = NULL;
     return 0;
 }
